package abc.ar.ast;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import polyglot.ast.Block;
import polyglot.ast.Local;
import polyglot.ast.MethodDecl;
import polyglot.ast.Node;
import polyglot.types.CodeInstance;
import polyglot.types.ConstructorInstance;
import polyglot.types.Context;
import polyglot.types.Flags;
import polyglot.types.MethodInstance;
import polyglot.types.SemanticException;
import polyglot.util.Position;
import polyglot.visit.AmbiguityRemover;
import polyglot.visit.NodeVisitor;
import abc.ar.ArOptions;
import abc.ar.types.ARContext;
import abc.ar.visit.NamedAspectMethods;
import abc.aspectj.ast.AJNodeFactory;
import abc.aspectj.ast.AdviceDecl_c;
import abc.aspectj.ast.AdviceFormal;
import abc.aspectj.ast.AdviceSpec;
import abc.aspectj.ast.Around;
import abc.aspectj.ast.Pointcut;
import abc.aspectj.types.AJTypeSystem;
import abc.aspectj.visit.AspectMethods;
import abc.weaving.aspectinfo.AbcFactory;
import abc.weaving.aspectinfo.Aspect;
import abc.weaving.aspectinfo.GlobalAspectInfo;
import abc.weaving.aspectinfo.MethodCategory;

public class NamedAdviceDecl_c extends AdviceDecl_c implements NamedAdviceDecl {

	protected boolean isNamedAroundAdvice;

	private NamedAdviceSpec namedSpec;

	public NamedAdviceDecl_c(Position pos, Flags flags, AdviceSpec spec, List throwTypes, Pointcut pc, Block body) {
		super(pos, flags, spec, throwTypes, pc, body);
		isNamedAroundAdvice = spec instanceof NamedAround;
		namedSpec = (NamedAdviceSpec) spec;
		boolean alwaysJoinPoint=ArOptions.v().getThisJointPointVisibility()!=ArOptions.THISJOINPOINT_NONE;
		hasJoinPoint=alwaysJoinPoint;
		hasJoinPointStaticPart=alwaysJoinPoint;
		hasEnclosingJoinPointStaticPart=alwaysJoinPoint;
	}

	public AdviceSpec getSpec() {
		return spec;
	}

	public void aspectMethodsEnter(AspectMethods visitor) {
		super.aspectMethodsEnter(visitor);
//		if (visitor instanceof NamedAspectMethods)
//			((NamedAspectMethods) visitor).pushNamedAroundFor(this);
	}

	public Node aspectMethodsLeave(AspectMethods visitor, AJNodeFactory nf, AJTypeSystem ts) {
		if (visitor instanceof NamedAspectMethods) {
//			MethodDecl md = ((NamedAspectMethods) visitor).namedAround();

			Node r = super.aspectMethodsLeave(visitor, nf, ts);

//			((NamedAspectMethods) visitor).popNamedAround();
//			if (md != null)
//				visitor.addMethod(md);

			return r;
		}

		return super.aspectMethodsLeave(visitor, nf, ts);
	}

	public MethodDecl namedAroundDecl(AJNodeFactory nf, AJTypeSystem ts) {
		// prototype method for named around method. created like the prototype
		// proceed message
//		if (isNamedAroundAdvice) {
//			TypeNode tn = (TypeNode) returnType().copy();
//			List formals = new LinkedList(formals());
//			Return ret;
//
//			if (tn.type() == ts.Void())
//				ret = nf.Return(position());
//			else {
//				Expr dummy = dummyVal(nf, tn.type());
//				ret = nf.Return(position(), dummy);
//			}
//
//			Block bl = nf.Block(position()).append(ret);
//			List thrws = new LinkedList();
//			String name = aroundSpec.name();// name given by user!
//			MethodDecl md = nf.MethodDecl(position(), Flags.PROTECTED,// not
//																		// public
//																		// static
//																		// final
//																		// as
//																		// proceed!
//					tn, name, formals, thrws, bl);
//			MethodInstance mi = ts.methodInstance(position(), methodInstance().container(), Flags.PROTECTED, tn.type(),
//					name, new ArrayList(methodInstance().formalTypes()), new ArrayList());
//			((ParsedClassType) methodInstance().container()).addMethod(mi);
//			md = md.methodInstance(mi);
//			
//			
//			aroundSpec.setNamedAround(md);
//			
//			return md;
//		} else
			return null;
	}

	   public void update(GlobalAspectInfo gai, Aspect current_aspect)
	    {
	        int lastpos = formals().size();
	        int jp = -1, jpsp = -1, ejp = -1;
	        if (hasEnclosingJoinPointStaticPart) ejp = --lastpos;
	        if (hasJoinPoint) jp = --lastpos;
	        if (hasJoinPointStaticPart) jpsp = --lastpos;

	        // Since the spec is not visited, we copy the (checked)
	        // return type node from the advice declaration
	        spec.setReturnType(returnType());
	        // And the return formal as well
	        if (retval != null) {
	            spec.setReturnVal(retval);
	        }
	        
	        List methods = new ArrayList();
	        for (Iterator procs = methodsInAdvice.iterator(); procs.hasNext(); ) {
	                CodeInstance ci = (CodeInstance) procs.next();
	                if (ci instanceof MethodInstance)
	                        methods.add(AbcFactory.MethodSig((MethodInstance)ci));
	                if (ci instanceof ConstructorInstance)
	                        methods.add(AbcFactory.MethodSig((ConstructorInstance)ci));
	        }

	        abc.weaving.aspectinfo.AdviceDecl ad =
	            new abc.ar.weaving.aspectinfo.NamedAdviceDecl
	            (spec.makeAIAdviceSpec(),
	             pc.makeAIPointcut(),
	             AbcFactory.MethodSig(this),
	             current_aspect,
	             jp, jpsp, ejp, methods,
	             position(),namedSpec.name());
	        gai.addAdviceDecl(ad);
	 

	        MethodCategory.register(this, MethodCategory.ADVICE_BODY);
	        if (spec instanceof Around) {
	            MethodCategory.register(((Around)spec).proceed(), MethodCategory.PROCEED);
	        }
	    }
//	public void update(GlobalAspectInfo gai, Aspect current_aspect) {
//		super.update(gai, current_aspect);
////        if (spec instanceof NamedAround) 
////            MethodCategory.register(((NamedAround)spec).namedAround(), MethodCategory.NORMAL);
//	}
	   
	   public Context enterScope(Context c) {
		   ((ARContext)c).pushNamedAdvice(this);
		return super.enterScope(c);
	}

	public AdviceFormal returnVal() {
		return retval;
	}
	
	
	 public NodeVisitor disambiguateEnter(AmbiguityRemover ar) throws SemanticException
	    {
		 //do not bypass retval, this is needed for creating the named method!
	        if (ar.kind() == AmbiguityRemover.SIGNATURES) {
	            if (body != null) {
	                Collection bp = new LinkedList();
	                bp.add(body);
	                bp.add(pc);
	                return ar.bypass(bp);
	            }
	        } else
	        	ar=(AmbiguityRemover)super.disambiguateEnter(ar);

	        return ar;
	    }
	 
	
}
