package abc.ar.weaving.weaver.around;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import soot.Local;
import soot.Modifier;
import soot.PatchingChain;
import soot.SootClass;
import soot.SootField;
import soot.SootMethod;
import soot.Type;
import soot.Value;
import soot.jimple.AssignStmt;
import soot.jimple.IdentityStmt;
import soot.jimple.Jimple;
import soot.jimple.JimpleBody;
import soot.jimple.ParameterRef;
import soot.jimple.Stmt;
import soot.jimple.VirtualInvokeExpr;
import soot.jimple.internal.JimpleLocal;
import soot.jimple.internal.JimpleLocalBox;
import soot.util.Chain;
import abc.soot.util.Restructure;
import abc.weaving.weaver.around.AdviceLocalClass;
import abc.weaving.weaver.around.AdviceLocalMethod;
import abc.weaving.weaver.around.AdviceMethod;

public class ARAdviceLocalMethod extends AdviceLocalMethod {
	protected String adviceName;

	protected boolean isNamed;

	protected SootMethod adviceNamedMethod;

	ARAdviceLocalMethod(AdviceLocalClass enclosingClass, AdviceMethod adviceMethod, SootMethod method, String adviceName) {
		super(enclosingClass, adviceMethod, method);
		this.adviceName = adviceName;
		this.isNamed = adviceName != null;

		//weave named advice in the constructor before the find proceed call
		Chain methodStatements = methodBody.getUnits().getNonPatchingChain();
		weaveNamedAdviceMethod(method, methodStatements);
	}


	private void weaveNamedAdviceMethod(SootMethod method, Chain methodStatements) {
		// //////////////////////////////////////////+
		if (isNamed) {
			// create named advice method with the code of the unnamed one
			SootMethod adviceNamedMethod;

			adviceNamedMethod = new SootMethod(adviceName, adviceMethod.originalAdviceFormalTypes, adviceMethod
					.getAdviceReturnType(), Modifier.PROTECTED);
			adviceNamedMethod.setActiveBody(method.getActiveBody());

			if (enclosingClass.sootClass.declaresMethod(adviceName, adviceMethod.originalAdviceFormalTypes,
					adviceMethod.getAdviceReturnType())) {
				SootMethod dummyMethod = enclosingClass.sootClass.getMethod(adviceName,
						adviceMethod.originalAdviceFormalTypes, adviceMethod.getAdviceReturnType());
				enclosingClass.sootClass.removeMethod(dummyMethod);
			}

			enclosingClass.sootClass.addMethod(adviceNamedMethod);
			this.adviceNamedMethod = adviceNamedMethod;

			// if unnamed advice method already has more parameters than the
			// original advice than load those params from fields (filled later)
			// instead of from parameters
			PatchingChain namedUnits = adviceNamedMethod.getActiveBody().getUnits();
			for (int p = adviceMethod.originalAdviceFormalTypes.size(); p < method.getParameterCount(); p++) {
				Object unit = namedUnits.getFirst();
				while (unit != null) {
					if (unit instanceof IdentityStmt) {
						IdentityStmt idstmt = (IdentityStmt) unit;
						if (idstmt.getRightOp() instanceof ParameterRef) {
							ParameterRef paramref = (ParameterRef) idstmt.getRightOp();
							if (paramref.getIndex() == p) {
								// create a field assignment and replace
								// identitystatement
								Local lThis = Restructure.getThisLocal(adviceNamedMethod);
								String localName=((JimpleLocal)idstmt.getLeftOp()).getName();
								SootField field = findFieldForParam(p,localName,method.getParameterType(p));
								AssignStmt fieldAssignment = Jimple.v().newAssignStmt(idstmt.getLeftOp(),
										Jimple.v().newInstanceFieldRef(lThis, field.makeRef()));
								namedUnits.insertBefore(fieldAssignment, unit);
								namedUnits.remove(unit);
								unit = fieldAssignment;
							}
						}
					}

					unit = namedUnits.getSuccOf(unit);
				}
			}

			// rewrite unnamed advice method

			JimpleBody newUnnamedBody = Jimple.v().newBody(method);
			{
				// declare this
				Chain locals = newUnnamedBody.getLocals();
				PatchingChain units = newUnnamedBody.getUnits();
				Local lThis = Jimple.v().newLocal("lthis", enclosingClass.sootClass.getType());
				locals.add(lThis);
				// initialize this
				units.add(Jimple.v().newIdentityStmt(lThis, Jimple.v().newThisRef(enclosingClass.sootClass.getType())));
				// declare all parametervariables &initialize all
				// parametervariables
				List params = new ArrayList();
				List assigns = new ArrayList();
				for (int p = 0; p < method.getParameterCount(); p++) {
					Type ptype = method.getParameterType(p);

					Local lP = Jimple.v().newLocal("lp" + p, ptype);
					locals.add(lP);

					units.add(Jimple.v().newIdentityStmt(lP, Jimple.v().newParameterRef(ptype, p)));
					// pass original parameters, but store additionl ones to
					// fields
					if (p < adviceMethod.originalAdviceFormalTypes.size())
						params.add(lP);
					else {
						assigns.add(Jimple.v().newAssignStmt(
								Jimple.v().newInstanceFieldRef(lThis, getFieldForParam(p).makeRef()), lP));
					}
				}
				units.addAll(assigns);
				// declare result
				Local lResult = Jimple.v().newLocal("lresult", adviceMethod.getAdviceReturnType());
				locals.add(lResult);
				// call named advice method
				// units.add();
				VirtualInvokeExpr invokeNamedMethod = Jimple.v().newVirtualInvokeExpr(lThis,
						adviceNamedMethod.makeRef(), params);
				units.add(Jimple.v().newAssignStmt((Value) lResult, (Value) invokeNamedMethod));
				// return result
				units.add(Jimple.v().newReturnStmt(lResult));
			}
			method.setActiveBody(newUnnamedBody);
		} else
			this.adviceNamedMethod = null;

		// //////////////////////////////////////////-
	}

	public Local addParameterToMethod(Type type, String suggestedName) {
		SootMethod unnamedMethod = sootProceedCallMethod;
		SootMethod namedMethod = adviceNamedMethod;
		if (isNamed) {
			// do not only extend one method but add parameter to a field and
			// reload
			// it where proceeding

			// add parameter and local to unnamed function
			Local newUnnamedLocal = Restructure.addParameterToMethod(unnamedMethod, type, suggestedName);

			// add field to aspect class
			SootClass cl = this.enclosingClass.sootClass;
			SootField newField = new SootField("_" + newUnnamedLocal.getName(), type, Modifier.PRIVATE);
			cl.addField(newField);

			// add field assignment to unnamed method
			Local lThis = Restructure.getThisLocal(unnamedMethod);
			AssignStmt fieldAssignment = Jimple.v().newAssignStmt(
					Jimple.v().newInstanceFieldRef(lThis, newField.makeRef()), newUnnamedLocal);

			PatchingChain unnamedUnits = unnamedMethod.getActiveBody().getUnits();
			Stmt insert = Restructure.findFirstRealStmtOrNop(unnamedMethod, unnamedUnits);
			unnamedUnits.insertBefore(fieldAssignment, insert);

			// add local to named method
			Chain namedLocals = namedMethod.getActiveBody().getLocals();
			Local newNamedLocal = Jimple.v().newLocal(newUnnamedLocal.getName(), newUnnamedLocal.getType());
			namedLocals.add(newNamedLocal);

			// assign local from field
			lThis = Restructure.getThisLocal(namedMethod);
			fieldAssignment = Jimple.v().newAssignStmt(newNamedLocal,
					Jimple.v().newInstanceFieldRef(lThis, newField.makeRef()));
			PatchingChain namedUnits = namedMethod.getActiveBody().getUnits();
			insert = Restructure.findFirstRealStmtOrNop(namedMethod, namedUnits);
			namedUnits.insertBefore(fieldAssignment, insert);

			return newNamedLocal;
		} else
			return super.addParameterToMethod(type, suggestedName);
	}
	
	
	private Map paramFieldMap;
	private SootField getFieldForParam(int p) {
		if( paramFieldMap==null)return null;
		SootField r=(SootField)paramFieldMap.get(new Integer(p));
		return r;
	}
	
	private Map nameFieldMap;
	private SootField findFieldForParam(int p, String name, Type type) {
		if (nameFieldMap==null) nameFieldMap=new HashMap();
		if (paramFieldMap==null) paramFieldMap=new HashMap();
		SootField r=(SootField)nameFieldMap.get(name);
		if (r!=null)
			return r;

		SootClass cl = this.enclosingClass.sootClass;
		r=cl.getField(name, type);
		
		nameFieldMap.put(name,r);
		paramFieldMap.put(new Integer(p),r);
		
		return r;
	}
}
